from yta_editor_nodes_gpu.processor import _NodeProcessorGPU
from typing import Union

import moderngl


class _VideoNodeProcessorGPU(_NodeProcessorGPU):
    """
    *Abstract class*

    *Singleton class*

    *For internal use only*

    Class to represent a node processor that uses GPU
    to transform the input but it is meant to video
    processing, so it implements a 'time' parameter to
    manipulate the frames according to that time
    moment.

    This class must be implemented by any processor
    that uses GPU to modify an input.
    """

    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        **kwargs
    ):
        """
        Provide all the variables you want to be initialized
        as uniforms at the begining for the global OpenGL
        animation in the `**kwargs`.

        The `output_size` is the size (width, height) of the
        texture that will be obtained as result. This size
        can be modified when processing a specific input, but
        be consider the cost of resources of modifying the 
        size, that will regenerate the output texture.
        """
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            **kwargs
        )

    def __reinit__(
        self,
        opengl_context: Union[moderngl.Context, None] = None,
        output_size: Union[tuple[int, int], None] = None,
        **kwargs
    ):
        super().__reinit__(
            opengl_context = opengl_context,
            output_size = output_size,
            **kwargs
        )
    
    def process(
        self,
        input: Union[moderngl.Texture, 'np.ndarray'],
        # TODO: What about this output size (?)
        output_size: Union[tuple[int, int], None] = None,
        t: float = 0.0,
        **kwargs
    # TODO: What about the output type (?)
    ) -> moderngl.Texture:
        """
        Process the provided `input` and transform it by
        using the code that is defined here according to
        the `t` time moment provided.

        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        return super().process(
            input = input,
            output_size = output_size,
            time = t,
            **kwargs
        )
    
# Specific implementations here below:
class BreathingFrameVideoNodeProcessorGPU(_VideoNodeProcessorGPU):
    """
    The frame but as if it was breathing.
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330
            uniform sampler2D base_texture;
            uniform float time;
            uniform float zoom;
            in vec2 v_uv;
            out vec4 output_color;
            // Use uniforms to be customizable

            void main() {
                // Dynamic zoom scaled with t
                float scale = 1.0 + zoom * sin(time * 2.0);
                vec2 center = vec2(0.5, 0.5);

                // Recalculate coords according to center
                vec2 uv = (v_uv - center) / scale + center;

                // Clamp to avoid artifacts
                uv = clamp(uv, 0.0, 1.0);

                output_color = texture(base_texture, uv);
            }
            '''
        )

    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        zoom: float = 0.05
        # TODO: Handle dynamic parameters with None
    ):
        """
        Provide all the variables you want to be initialized
        as uniforms at the begining for the global OpenGL
        animation in the `**kwargs`.

        The `output_size` is the size (width, height) of the
        texture that will be obtained as result. This size
        can be modified when processing a specific input, but
        be consider the cost of resources of modifying the 
        size, that will regenerate the output texture.
        """
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            # Uniforms
            zoom = zoom
        )

    def __reinit__(
        self,
        opengl_context: Union[moderngl.Context, None] = None,
        output_size: Union[tuple[int, int], None] = None,
        zoom: Union[float, None] = None
        # TODO: Handle dynamic parameters with None
    ):
        super().__reinit__(
            opengl_context = opengl_context,
            output_size = output_size,
            zoom = zoom
        )
    
class WavingFramesVideoNodeProcessorGPU(_VideoNodeProcessorGPU):
    """
    The GPU processor of the video frames that are
    transformed into a wave.

    TODO: Explain this better.
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330
            uniform sampler2D base_texture;
            uniform float time;
            uniform float amplitude;
            uniform float frequency;
            uniform float speed;
            uniform bool do_use_transparent_pixels;
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                float wave = sin(v_uv.x * frequency + time * speed) * amplitude;
                vec2 uv = vec2(v_uv.x, v_uv.y + wave);

                // Si el UV se sale del rango, devolvemos transparencia
                if ((uv.y < 0.0 || uv.y > 1.0) && do_use_transparent_pixels) {
                    output_color = vec4(0.0, 0.0, 0.0, 0.0);
                } else {
                    output_color = texture(base_texture, uv);
                }
            }
            '''
        )
    
    # TODO: This below doesn't use transparent pixels but
    # the other pixels of the image instead
    """
    void main() {
        float wave = sin(v_uv.x * frequency + time * speed) * amplitude;
        vec2 uv = vec2(v_uv.x, v_uv.y + wave);
        output_color = texture(base_texture, uv);
    }
    """
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        # Uniforms
        amplitude: float = 0.05,
        frequency: float = 10.0,
        speed: float = 2.0,
        do_use_transparent_pixels: bool = False
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            # Uniforms
            amplitude = amplitude,
            frequency = frequency,
            speed = speed,
            do_use_transparent_pixels = do_use_transparent_pixels
        )

    def __reinit__(
        self,
        opengl_context: Union[moderngl.Context, None] = None,
        output_size: Union[tuple[int, int], None] = None,
        # Uniforms
        amplitude: Union[float, None] = None,
        frequency: Union[float, None] = None,
        speed: Union[float, None] = None,
        do_use_transparent_pixels: Union[bool, None] = None
    ):
        super().__reinit__(
            opengl_context = opengl_context,
            output_size = output_size,
            amplitude = amplitude,
            frequency = frequency,
            speed = speed,
            do_use_transparent_pixels = do_use_transparent_pixels
        )



"""
This entire module is working with OpenGL.

We don't name it 'blend' instead of 'process' because
the GPU classes have the 'process' name due to the
inheritance from OpenGL classes, and we want to avoid
misunderstandings.

(!) IMPORTANT:
In all our blenders we need to include a platform
called `mix_weight` to determine the percentage of 
effect we want to apply to the output result:
- `uniform float mix_weight;  // 0.0 → only base, 1.0 → only overlay`
- `output_color = mix(base_color, overlay_color, mix_weight);`
"""
from yta_video_opengl.abstract import _OpenGLBase
from yta_editor_nodes_gpu.processor.blender.utils import _validate_inputs_and_mix_weights
from yta_video_opengl.texture.utils import TextureUtils
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from typing import Union

import numpy as np
import moderngl


# TODO: I think this one should inherit from
# '_NodeProcessorGPU'
class _BlenderGPU(_OpenGLBase):
    """
    *Singleton class*
    
    *For internal use only*

    *This class has to be inherited*

    Class to represent a blender that uses GPU to
    blend the inputs. Base class for an OpenGL
    blender, that will be inherited by the specific
    implementations.

    The basic class of a blender to blend frames as
    opengl textures. This blender will process the
    inputs as textures and will generate also a
    texture as the output.

    Blender results can be chained and the result
    from one node can be applied on another node.

    This blender is able to work acting as an
    opacity blender, but will be used as the base
    of the classes that we will implement actually.

    The `opacity` value must be set as an uniform
    each time we are processing the 2 textures.
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        """
        The code of the fragment shader.
        """
        return (
            '''
            #version 330
            uniform sampler2D base_texture;
            uniform sampler2D overlay_texture;
            uniform float mix_weight;  // 0.0 → only base, 1.0 → only overlay
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                vec4 base_color = texture(base_texture, v_uv);
                vec4 overlay_color = texture(overlay_texture, v_uv);

                // Apply global mix intensity
                output_color = mix(base_color, overlay_color, mix_weight);
            }
            '''
        )
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        mix_weight: float = 1.0,
        **kwargs
    ):
        ParameterValidator.validate_mandatory_number_between('mix_weight', mix_weight, 0.0, 1.0)

        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            mix_weight = mix_weight,
            **kwargs
        )
    
    # TODO: Overwrite this method
    def _prepare_input_textures(
        self
    ) -> '_BlenderGPU':
        """
        *For internal use only*

        Set the input texture variables and handlers
        we need to manage this. This method has to be
        called only once, just to set the slot for 
        the different textures we will use (and are
        registered as textures in the shader).
        """
        self.textures.add('base_texture', 0)
        self.textures.add('overlay_texture', 1)

        return self
    
    # TODO: Overwrite this method
    def process(
        self,
        base_input: Union[moderngl.Texture, np.ndarray],
        overlay_input: Union[moderngl.Texture, np.ndarray],
        output_size: Union[tuple[int, int], None] = None,
        mix_weight: float = 1.0,
        **kwargs
    ) -> moderngl.Texture:
        """
        Validate the parameters, set the textures map, process
        the mix by applying the blender logic, and return the
        result but only applied as much as the `mix_weight` 
        parameter is indicating.

        Apply the shader to the given 'base_input'
        and 'overlay_input', that must be frames or
        textures, and return the new resulting
        texture.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        ParameterValidator.validate_mandatory_instance_of('base_input', base_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_instance_of('overlay_input', overlay_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_number_between('mix_weight', mix_weight, 0.0, 1.0)

        if mix_weight == 0.0:
            # If the mix_weight is 0.0 we don't want to affect
            # the base so we don't even need to process the
            # mix but return the original base input as a
            # Texture
            return (
                base_input
                if PythonValidator.is_instance_of(base_input, moderngl.Texture) else
                TextureUtils.numpy_to_texture(
                    input = base_input,
                    opengl_context = self.context
                )
            )

        textures_map = {
            'base_texture': base_input,
            'overlay_texture': overlay_input
        }

        return self._process_common(
            textures_map = textures_map,
            output_size = output_size,
            mix_weight = mix_weight,
            **kwargs
        )
    
    # TODO: This method can change according to the way
    # we need to process the inputs, that could be
    # different than by pairs
    def process_multiple_inputs(
        self,
        inputs: list[Union[np.ndarray, moderngl.Texture]],
        mix_weights: Union[list[float], float] = 1.0,
        output_size: Union[tuple[int, int], None] = None,
    ) -> moderngl.Texture:
        """
        Blend all the `inputs` provided, one after another,
        applying the `mix_weights` provided, and forcing the
        result to the `dtype` if provided.

        The `mix_weights` can be a single float value, that 
        will be used for all the mixings, or a list of as
        many float values as `inputs` received, to be 
        applied individually to each mixing.
        """
        _validate_inputs_and_mix_weights(
            inputs = inputs,
            mix_weights = mix_weights
        )
        
        # We process all the 'inputs' as 'base' and 'overlay'
        # and accumulate the result

        # Use the first one as the base
        output = inputs[0]

        # TODO: How do we handle the additional parameters that
        # could be an array? Maybe if it is an array, check that
        # the number of elements is the same as the number of
        # inputs, and if a single value just use it...

        for i in range(1, len(inputs)):
            overlay_input = inputs[i]
            mix_weight = mix_weights[i]

            output = self.process(
                base_input = output,
                overlay_input = overlay_input,
                output_size = output_size,
                mix_weight = mix_weight,
            )

        return output
    
# Specific implementations below
class MixBlenderGPU(_BlenderGPU):
    """
    Blender to blend 2 textures by using a mix float
    parameter that determines the amount of the result
    we want to mix with the input.
    """
    pass

class AlphaBlenderGPU(_BlenderGPU):
    """
    Blender to blend 2 textures by using the most common
    blending method, which is the alpha.

    This blender will use the alpha channel of the 
    overlay input, multiplied by the `blend_strength`
    parameter provided, to use it as the mixer factor
    between the base and the overlay inputs.
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        """
        The code of the fragment shader.
        """
        # TODO: What if no 'overlay_color.a' (?)
        return (
            '''
            #version 330
            uniform sampler2D base_texture;
            uniform sampler2D overlay_texture;
            uniform float mix_weight;  // 0.0 → only base, 1.0 → only overlay
            uniform float blend_strength;
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                vec4 base_color = texture(base_texture, v_uv);
                vec4 overlay_color = texture(overlay_texture, v_uv);

                // Use alpha channel of overlay if available
                float alpha = overlay_color.a * blend_strength;

                // Classic blending
                output_color = mix(base_color, overlay_color, alpha);

                // Apply global mix intensity
                output_color = mix(base_color, output_color, mix_weight);
            }
            '''
        )
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int]
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size
        )

    def process(
        self,
        base_input: Union[moderngl.Texture, np.ndarray],
        overlay_input: Union[moderngl.Texture, np.ndarray],
        output_size: Union[tuple[int, int], None] = None,
        mix_weight: float = 1.0,
        blend_strength: float = 0.5
    ) -> moderngl.Texture:
        """
        Validate the parameters, set the textures map, process
        the mix by applying the blender logic, and return the
        result but only applied as much as the `mix_weight` 
        parameter is indicating.

        Apply the shader to the given 'base_input'
        and 'overlay_input', that must be frames or
        textures, and return the new resulting
        texture.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        ParameterValidator.validate_mandatory_instance_of('base_input', base_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_instance_of('overlay_input', overlay_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_number_between('mix_weight', mix_weight, 0.0, 1.0)
        ParameterValidator.validate_mandatory_number_between('blend_strength', blend_strength, 0.0, 1.0)

        if mix_weight == 0.0:
            # If the mix_weight is 0.0 we don't want to affect
            # the base so we don't even need to process the
            # mix but return the original base input as a
            # Texture
            return (
                base_input
                if PythonValidator.is_instance_of(base_input, moderngl.Texture) else
                TextureUtils.numpy_to_texture(
                    input = base_input,
                    opengl_context = self.context
                )
            )

        textures_map = {
            'base_texture': base_input,
            'overlay_texture': overlay_input
        }

        return self._process_common(
            textures_map = textures_map,
            output_size = output_size,
            mix_weight = mix_weight,
            blend_strength = blend_strength
        )

class AddBlenderGPU(_BlenderGPU):
    """
    Blender to blend 2 textures by adding their values.

    This blender will increase the brightness by
    combining the colors of the base and the overlay
    inputs, using the overlay as much as the 
    `stregth` parameter is indicating.
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        """
        The code of the fragment shader.
        """
        return (
            '''
            #version 330
            uniform sampler2D base_texture;
            uniform sampler2D overlay_texture;
            uniform float mix_weight;  // 0.0 → only base, 1.0 → only overlay
            uniform float strength;  // 1.0 = full additive, 0.5 = subtle
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                vec4 base_color = texture(base_texture, v_uv);
                vec4 overlay_color = texture(overlay_texture, v_uv);

                vec4 result = base_color + overlay_color * strength;

                // Clamp to avoid overflow (values > 1.0)
                output_color = clamp(result, 0.0, 1.0);

                // Apply global mix intensity
                output_color = mix(base_color, output_color, mix_weight);
            }
            '''
        )
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
        )

    def process(
        self,
        base_input: Union[moderngl.Texture, np.ndarray],
        overlay_input: Union[moderngl.Texture, np.ndarray],
        output_size: Union[tuple[int, int], None] = None,
        mix_weight: float = 1.0,
        strength: float = 0.5
    ) -> moderngl.Texture:
        """
        Validate the parameters, set the textures map, process
        the mix by applying the blender logic, and return the
        result but only applied as much as the `mix_weight` 
        parameter is indicating.

        Apply the shader to the given 'base_input'
        and 'overlay_input', that must be frames or
        textures, and return the new resulting
        texture.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        ParameterValidator.validate_mandatory_instance_of('base_input', base_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_instance_of('overlay_input', overlay_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_number_between('mix_weight', mix_weight, 0.0, 1.0)
        ParameterValidator.validate_mandatory_number_between('strength', strength, 0.0, 1.0)

        if mix_weight == 0.0:
            # If the mix_weight is 0.0 we don't want to affect
            # the base so we don't even need to process the
            # mix but return the original base input as a
            # Texture
            return (
                base_input
                if PythonValidator.is_instance_of(base_input, moderngl.Texture) else
                TextureUtils.numpy_to_texture(
                    input = base_input,
                    opengl_context = self.context
                )
            )

        textures_map = {
            'base_texture': base_input,
            'overlay_texture': overlay_input
        }

        return self._process_common(
            textures_map = textures_map,
            output_size = output_size,
            mix_weight = mix_weight,
            strength = strength
        )

# TODO: Create more



"""
TODO: This module must be moved.

These classes are to store the trajectory that
will be managed by the timeline (or other 
instance related to this) and pass the values
to render to the node that is capable of it. The
trajectory will not be stored by the node.
"""
from yta_validation.parameter import ParameterValidator
from dataclasses import dataclass

import numpy as np

        
@dataclass
class TrajectoryCoordinate:
    """
    Dataclass to store a trajectory coordinate, that is
    a position (x, y), the rotation of the object and some
    aditional parameters.
    """

    def __init__(
        self,
        position: tuple[int, int],
        rotation: int = 0,
        additional_params: dict = {}
    ):
        """
        - The `position` parameter must be an (x, y) position
        in pixels.
        - The `rotation` must be an int in the range [0, 360].
        """
        self.position: tuple[int, int] = position
        """
        The position (x, y) of the center of the item that
        will be moving doing this trajectory.
        """
        self.rotation: int = rotation
        """
        The rotation of the element whose center will be
        positioned in the given `position`.
        """
        self._additional_params: dict = additional_params
        """
        Some additional parameters to include more information
        of the movement.
        """

@dataclass
class TrajectoryCoordinateWithT(TrajectoryCoordinate):
    """
    Dataclass to store a trajectory coordinate, that is
    a position (x, y), the rotation of the object, a time
    moment and some aditional parameters.
    """

    def __init__(
        self,
        position: tuple[int, int],
        t: float,
        rotation: int = 0,
        additional_params: dict = {}
    ):
        super().__init__(
            position = position,
            rotation = rotation,
            additional_params = additional_params
        )

        self.t: float = t
        """
        The time moment in which the center of the object
        must be placed in this position.
        """

@dataclass
class Trajectory:
    """
    Dataclass to store a set of consecutive coordinates
    that build a trajectory for some object. It includes
    the rotation and also can have some aditional
    parameters.
    """

    @property
    def coordinates(
        self
    ) -> list[TrajectoryCoordinate]:
        """
        The coordinates of the trajectory, ordered as they
        were added.
        """
        return self._coordinates

    def __init__(
        self,
        coordinates: list[TrajectoryCoordinate]
    ):
        ParameterValidator.validate_mandatory_list_of_these_instances('coordinates', coordinates, TrajectoryCoordinate)

        self._coordinates: list[TrajectoryCoordinate] = coordinates
        """
        The coordinates in which the center of the object
        that follows the trajectory will be placed.
        """

    def add_coordinate(
        self,
        coordinate: TrajectoryCoordinate
    ) -> 'Trajectory':
        """
        Add a new coordinate to the list.
        """
        ParameterValidator.validate_mandatory_instance_of('coordinate', coordinate, TrajectoryCoordinate)

        self._coordinates.append(coordinate)

        return self
    
@dataclass
class TrajectoryWithT(Trajectory):
    """
    Dataclass to store a set of consecutive coordinates
    that build a trajectory for some object. It includes
    the rotation, the time moment and also can have some
    aditional parameters.
    """
    
    @property
    def min_t(
        self
    ) -> float:
        """
        Get the min `t` of the coordinates of this trajectory.
        """
        return self.coordinates[0].t
    
    @property
    def max_t(
        self
    ) -> float:
        """
        Get the max `t` of the coordinates of this trajectory.
        """
        return self.coordinates[-1].t
    
    @property
    def coordinates(
        self
    ) -> list[TrajectoryCoordinateWithT]:
        """
        The coordinates but ordered by the `t` time moment.
        """
        if not hasattr(self, '_coordinates_ordered_by_t'):
            self._coordinates_ordered_by_t = sorted(self._coordinates, key = lambda coordinate: coordinate.t)

        return self._coordinates_ordered_by_t

    def __init__(
        self,
        coordinates: list[TrajectoryCoordinateWithT]
    ):
        ParameterValidator.validate_mandatory_list_of_these_instances('coordinates', coordinates, TrajectoryCoordinateWithT)

        self._coordinates: list[TrajectoryCoordinateWithT] = coordinates
        """
        The coordinates in which the center of the object
        that follows the trajectory will be placed.
        """

    def add_coordinate(
        self,
        coordinate: TrajectoryCoordinateWithT
    ) -> 'TrajectoryWithT':
        """
        Add a new coordinate to the list.
        """
        ParameterValidator.validate_mandatory_instance_of('coordinate', coordinate, TrajectoryCoordinateWithT)

        self._coordinates.append(coordinate)
        self._coordinates_ordered_by_t = None

        return self

    def get_position_at(
        self,
        t: float
    ) -> tuple[int, int]:
        """
        Get a coordinate for the `t` time moment provided
        that includes the position (interpolated from the
        ones we actually have in the trajectory) and maybe
        some additional parameters if existing.
        """
        if t <= self.min_t:
            return self._coordinates[0]
        if t >= self.max_t:
            return self._coordinates[-1]

        # Extraer the times to be able to search
        times = [
            coordinate.t
            for coordinate in self.coordinates
        ]

        i = np.searchsorted(times, t) - 1

        # This is the factor to interpolate
        interpolation_factor = (t - self.coordinates[i].t) / (self.coordinates[i + 1].t - self.coordinates[i].t)

        return TrajectoryCoordinateWithT(
            # Interpolated position
            position = (1 - interpolation_factor) * self.coordinates[i].position + interpolation_factor * self.coordinates[i + 1].position,
            t = t,
            # Interpolated rotation
            rotation = (1 - interpolation_factor) * self.coordinates[i].rotation + interpolation_factor * self.coordinates[i + 1].rotation,
            # TODO: By now we are providing the additional
            # parameters of the previous coordinate
            additional_params = self.coordinates[i]._additional_params
        )


"""
Experimental module. The effects here
have not been tested completely or the
result is not as good as it should be
to be considered a definitive effect.

TODO: This implementation is old and
now we are using other classes, but
the shader and the variables are the
same and that is why we keep the code
here, to implement it later... (GPU)
"""
from yta_video_opengl.abstract import _OpenGLBase
from typing import Union

import numpy as np
import moderngl
import math


class BreathingFrame(_OpenGLBase):
    """
    The frame but as if it was breathing.
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330
            uniform sampler2D tex;
            uniform float time;
            in vec2 v_text;
            out vec4 f_color;
            // Use uniforms to be customizable

            void main() {
                // Dynamic zoom scaled with t
                float scale = 1.0 + 0.05 * sin(time * 2.0);  // 5% de zoom
                vec2 center = vec2(0.5, 0.5);

                // Recalculate coords according to center
                vec2 uv = (v_text - center) / scale + center;

                // Clamp to avoid artifacts
                uv = clamp(uv, 0.0, 1.0);

                f_color = texture(tex, uv);
            }
            '''
        )
    
    def process(
        self,
        input: Union[moderngl.Texture, 'np.ndarray'],
        t: float = 0.0,
    ) -> moderngl.Texture:
        """
        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        return super().process(
            input = input,
            time = t
        )
    
class HandheldFrame(_OpenGLBase):
    """
    The frame but as if it was being recorded by
    someone holding a camera, that is not 100%
    stable.
    """

    @property
    def vertex_shader(
        self
    ) -> str:
        return (
            '''
            #version 330
            in vec2 in_vert;
            in vec2 in_texcoord;
            out vec2 v_text;

            uniform mat3 transform;

            void main() {
                vec3 pos = vec3(in_vert, 1.0);
                pos = transform * pos;
                gl_Position = vec4(pos.xy, 0.0, 1.0);
                v_text = in_texcoord;
            }
            '''
        )
    
    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330
            uniform sampler2D tex;
            in vec2 v_text;
            out vec4 f_color;

            void main() {
                f_color = texture(tex, v_text);
            }
            '''
        )
    
    def _handheld_matrix(
        self,
        t
    ):
        # Rotación más notoria
        angle = self._smooth_noise(t, freq=0.8, scale=0.05)  # antes 0.02

        # Traslaciones más grandes
        tx = self._smooth_noise(t, freq=1.1, scale=0.04)     # antes 0.015
        ty = self._smooth_noise(t, freq=1.4, scale=0.04)

        # Zoom más agresivo
        zoom = 1.0 + self._smooth_noise(t, freq=0.5, scale=0.06)  # antes 0.02

        cos_a, sin_a = math.cos(angle), math.sin(angle)

        return np.array([
            [ cos_a * zoom, -sin_a * zoom, tx],
            [ sin_a * zoom,  cos_a * zoom, ty],
            [ 0.0,           0.0,          1.0]
        ], dtype="f4")

    def _smooth_noise(
        self,
        t,
        freq = 1.5,
        scale = 1.0
    ):
        """
        Small noise by using sin and cos mixed.
        """
        return (
            math.sin(t * freq) +
            0.5 * math.cos(t * freq * 0.5 + 1.7) +
            0.25 * math.sin(t * freq * 0.25 + 2.5)
        ) * scale
    
    def process(
        self,
        input: Union[moderngl.Texture, np.ndarray],
        t: float = 0.0,
    ) -> moderngl.Texture:
        """
        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        return super().process(
            input = input,
            # TODO: It was 'set_mat' previously
            # self._handheld_matrix(t).tobytes()
            transform = self._handheld_matrix(t)
        )
    
class OrbitingFrame(_OpenGLBase):
    """
    The frame but orbiting around the camera.
    """

    @property
    def vertex_shader(
        self
    ) -> str:
        return (
            '''
            #version 330

            in vec2 in_vert;
            in vec2 in_texcoord;

            out vec2 v_uv;

            uniform mat4 mvp;   // Model-View-Projection matrix

            void main() {
                v_uv = in_texcoord;
                // El quad está en XY, lo pasamos a XYZ con z=0
                vec4 pos = vec4(in_vert, 0.0, 1.0);
                gl_Position = mvp * pos;
            }
            '''
        )
    
    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330

            uniform sampler2D tex;
            in vec2 v_uv;
            out vec4 f_color;

            void main() {
                f_color = texture(tex, v_uv);
            }
            '''
        )
    
    def _perspective(
        self,
        fov_y_rad,
        aspect,
        near,
        far
    ):
        f = 1.0 / np.tan(fov_y_rad / 2.0)
        m = np.zeros((4,4), dtype='f4')
        m[0,0] = f / aspect
        m[1,1] = f
        m[2,2] = (far + near) / (near - far)
        m[2,3] = (2 * far * near) / (near - far)
        m[3,2] = -1.0

        return m
    
    def _look_at(
        self,
        eye,
        target,
        up = (0, 1, 0)
    ):
        eye = np.array(eye, dtype='f4')
        target = np.array(target, dtype='f4')
        up = np.array(up, dtype='f4')

        f = target - eye
        f = f / np.linalg.norm(f)
        s = np.cross(f, up)
        s = s / np.linalg.norm(s)
        u = np.cross(s, f)

        m = np.eye(4, dtype='f4')
        m[0,0:3] = s
        m[1,0:3] = u
        m[2,0:3] = -f
        m[0,3] = -np.dot(s, eye)
        m[1,3] = -np.dot(u, eye)
        m[2,3] =  np.dot(f, eye)

        return m

    def _translate(
        self,
        x,
        y,
        z
    ):
        m = np.eye(4, dtype='f4')
        m[0,3] = x
        m[1,3] = y
        m[2,3] = z

        return m

    def _rotate_y(
        self,
        angle
    ):
        c, s = np.cos(angle), np.sin(angle)
        m = np.eye(4, dtype='f4')
        m[0,0], m[0,2] =  c,  s
        m[2,0], m[2,2] = -s,  c

        return m
    
    def _scale_uniform(
        self,
        k
    ):
        m = np.eye(4, dtype='f4')
        m[0,0] = m[1,1] = m[2,2] = k

        return m
    
    def _carousel_mvp(
        self,
        t,
        *,
        aspect,
        fov_deg = 60.0,
        radius = 4.0,
        center_z = -6.0,
        speed = 1.0,
        face_center_strength = 1.0,
        extra_scale = 1.0
    ):
        """
        t: tiempo en segundos
        aspect: width/height del framebuffer
        radius: radio en XZ
        center_z: desplaza el carrusel entero hacia -Z para que esté frente a cámara
        speed: velocidad angular
        face_center_strength: 1.0 = panel mira al centro; 0.0 = no gira con la órbita
        """

        # Proyección y vista (cámara en el origen mirando hacia -Z)
        proj = self._perspective(np.radians(fov_deg), aspect, 0.1, 100.0)
        view = np.eye(4, dtype='f4')  # o look_at((0,0,0), (0,0,-1))

        # Ángulo de órbita (elige el offset para que "entre" por la izquierda)
        theta = speed * t - np.pi * 0.5

        # Órbita en XZ con el centro desplazado a center_z
        # x = radius * np.cos(theta)
        # z = radius * np.sin(theta) + center_z
        x = radius * np.cos(theta)
        z = (radius * 0.2) * np.sin(theta) + center_z

        # Yaw para que el panel apunte al centro (0,0,center_z)
        # El vector desde panel -> centro es (-x, 0, center_z - z)
        yaw_to_center = np.arctan2(-x, (center_z - z))  # atan2(X, Z)
        yaw = face_center_strength * yaw_to_center

        model = self._translate(x, 0.0, z) @ self._rotate_y(yaw) @ self._scale_uniform(extra_scale)

        # ¡IMPORTANTE! OpenGL espera column-major: transponemos al escribir
        mvp = proj @ view @ model
        
        return mvp
    
    def process(
        self,
        input: Union[moderngl.Texture, np.ndarray],
        t: float = 0.0,
    ) -> moderngl.Texture:
        """
        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        aspect = self.size[0] / self.size[1]
        mvp = self._carousel_mvp(t, aspect=aspect, radius=4.0, center_z=-4.0, speed=1.2, face_center_strength=1.0, extra_scale = 1.0)

        return super().process(
            input = input,
            # TODO: It was 'set_mat' previously
            # mvp.T.tobytes()
            mvp = mvp.T
        )
    
class RotatingInCenterFrame(_OpenGLBase):
    """
    The frame but orbiting around the camera.
    """

    @property
    def vertex_shader(
        self
    ) -> str:
        return (
            '''
            #version 330

            in vec2 in_vert;
            in vec2 in_texcoord;
            out vec2 v_uv;

            uniform float time;
            uniform float speed;

            void main() {
                v_uv = in_texcoord;

                // Rotación alrededor del eje Y
                float angle = time * speed;              // puedes usar time directamente, o time * speed
                float cosA = cos(angle);
                float sinA = sin(angle);

                // Convertimos el quad a 3D (x, y, z)
                vec3 pos = vec3(in_vert.xy, 0.0);

                // Rotación Y
                mat3 rotY = mat3(
                    cosA, 0.0, sinA,
                    0.0 , 1.0, 0.0,
                -sinA, 0.0, cosA
                );

                pos = rotY * pos;

                gl_Position = vec4(pos, 1.0);
            }
            '''
        )
    
    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330

            in vec2 v_uv;
            out vec4 f_color;

            uniform sampler2D tex;

            void main() {
                f_color = texture(tex, v_uv);
            }
            '''
        )
    
    def __init__(
        self,
        size,
        opengl_context = None,
        speed: float = 30
    ):
        super().__init__(
            opengl_context = opengl_context,
            size = size,
            speed = speed,
        )

    def process(
        self,
        input: Union[moderngl.Texture, np.ndarray],
        t: float = 0.0,
    ) -> moderngl.Texture:
        """
        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        return super().process(
            input = input,
            time = t
        )
    
class StrangeTvFrame(_OpenGLBase):
    """
    Nice effect like a tv screen or something...
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330

            uniform sampler2D tex;
            uniform float time;

            // ---- Parámetros principales (ajústalos en runtime) ----
            uniform float aberr_strength;  // 0..3   (fuerza del RGB split radial)
            uniform float barrel_k;        // -0.5..0.5  (distorsión de lente; positivo = barrel)
            uniform float blur_radius;     // 0..0.02 (radio de motion blur en UV)
            uniform float blur_angle;      // en radianes (dirección del arrastre)
            uniform int   blur_samples;    // 4..24  (taps del blur)
            uniform float vignette_strength; // 0..2
            uniform float grain_amount;    // 0..0.1
            uniform float flicker_amount;  // 0..0.2
            uniform float scanline_amount; // 0..0.2

            in vec2 v_uv;
            out vec4 f_color;

            // --- helpers ---
            float rand(vec2 co){
                return fract(sin(dot(co, vec2(12.9898,78.233))) * 43758.5453);
            }

            // Barrel distortion (simple, k>0 curva hacia fuera)
            vec2 barrel(vec2 uv, float k){
                // map to [-1,1]
                vec2 p = uv * 2.0 - 1.0;
                float r2 = dot(p, p);
                p *= (1.0 + k * r2);
                // back to [0,1]
                return p * 0.5 + 0.5;
            }

            // Aberración cromática radial
            vec3 sample_chromatic(sampler2D t, vec2 uv, vec2 center, float strength){
                // Offset radial según distancia al centro
                vec2 d = uv - center;
                float r = length(d);
                vec2 dir = (r > 1e-5) ? d / r : vec2(0.0);
                // Cada canal se desplaza un poco distinto
                float s = strength * r * 0.005; // escala fina
                float sr = s * 1.0;
                float sg = s * 0.5;
                float sb = s * -0.5; // azul hacia dentro para contraste

                float rC = texture(t, uv + dir * sr).r;
                float gC = texture(t, uv + dir * sg).g;
                float bC = texture(t, uv + dir * sb).b;
                return vec3(rC, gC, bC);
            }

            void main(){
                vec2 uv = v_uv;
                vec2 center = vec2(0.5, 0.5);

                // Lente (barrel/pincushion)
                uv = barrel(uv, barrel_k);

                // Early out si nos salimos mucho (fade de bordes)
                vec2 uv_clamped = clamp(uv, 0.0, 1.0);
                float edge = smoothstep(0.0, 0.02, 1.0 - max(max(-uv.x, uv.x-1.0), max(-uv.y, uv.y-1.0)));

                // Dirección del motion blur
                vec2 dir = vec2(cos(blur_angle), sin(blur_angle));
                // Pequeña variación temporal para que “respire”
                float jitter = (sin(time * 13.0) * 0.5 + 0.5) * 0.4 + 0.6;

                // Acumulación de blur con pesos
                vec3 acc = vec3(0.0);
                float wsum = 0.0;

                int N = max(1, blur_samples);
                for(int i = 0; i < 64; ++i){         // hard cap de seguridad
                    if(i >= N) break;
                    // t de -1..1 distribuye muestras a ambos lados
                    float fi = float(i);
                    float t = (fi / float(N - 1)) * 2.0 - 1.0;

                    // curva de pesos (gauss approx)
                    float w = exp(-t*t * 2.5);
                    // offset base
                    vec2 ofs = dir * t * blur_radius * jitter;

                    // micro-jitter por muestra para romper banding
                    ofs += vec2(rand(uv + fi)*0.0005, rand(uv + fi + 3.14)*0.0005) * blur_radius;

                    // muestreo con aberración cromática
                    vec3 c = sample_chromatic(tex, uv + ofs, center, aberr_strength);

                    acc += c * w;
                    wsum += w;
                }
                vec3 col = acc / max(wsum, 1e-6);

                // Scanlines + flicker
                float scan = 1.0 - scanline_amount * (0.5 + 0.5 * sin((uv.y + time*1.7)*3.14159*480.0));
                float flick = 1.0 + flicker_amount * (sin(time*60.0 + uv.x*10.0) * 0.5 + 0.5);
                col *= scan * flick;

                // Vignette (radial)
                float r = distance(uv, center);
                float vig = 1.0 - smoothstep(0.7, 1.0, r * (1.0 + 0.5*vignette_strength));
                col *= mix(1.0, vig, vignette_strength);

                // Grano
                float g = (rand(uv * (time*37.0 + 1.0)) - 0.5) * 2.0 * grain_amount;
                col += g;

                // Fade de bordes por clamp/warp
                col *= edge;

                f_color = vec4(col, 1.0);
            }
            '''
        )
    
    def __init__(
        self,
        size,
        opengl_context = None,
        aberr_strength: float = 1.5,
        barrel_k: float = 0.08,
        blur_radius: float = 0.006,
        blur_angle: float = 0.0, # (0 = horizontal, 1.57 ≈ vertical)
        blur_samples: int = 12,
        vignette_strength: float = 0.8,
        grain_amount: float = 0.02,
        flicker_amount: float = 0.05,
        scanline_amount: float = 0.05
    ):
        super().__init__(
            opengl_context = opengl_context,
            size = size,
            aberr_strength = aberr_strength,
            barrel_k = barrel_k,
            blur_radius = blur_radius,
            blur_angle = blur_angle,
            blur_samples = blur_samples,
            vignette_strength = vignette_strength,
            grain_amount = grain_amount,
            flicker_amount = flicker_amount,
            scanline_amount = scanline_amount
        )

    def process(
        self,
        input: Union[moderngl.Texture, np.ndarray],
        t: float = 0.0,
    ) -> moderngl.Texture:
        """
        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        return super().process(
            input = input,
            time = t
        )
    
class GlitchRgbFrame(_OpenGLBase):
    """
    Nice effect like a tv screen or something...
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330

            // ----------- Fragment Shader -----------
            uniform sampler2D tex;
            uniform float time;

            // Intensidades del efecto
            uniform float amp;      // amplitud de distorsión
            uniform float freq;     // frecuencia de la onda
            uniform float glitchAmp; // fuerza del glitch
            uniform float glitchSpeed;

            in vec2 v_uv;
            out vec4 f_color;

            void main() {
                // Distorsión sinusoidal en Y
                float wave = sin(v_uv.x * freq + time * 2.0) * amp;

                // Pequeño desplazamiento aleatorio (shake)
                float shakeX = (fract(sin(time * 12.9898) * 43758.5453) - 0.5) * 0.01;
                float shakeY = (fract(sin(time * 78.233) * 12345.6789) - 0.5) * 0.01;

                // Coordenadas base con distorsión
                vec2 uv = vec2(v_uv.x + shakeX, v_uv.y + wave + shakeY);

                // Glitch con separación RGB
                float glitch = sin(time * glitchSpeed) * glitchAmp;
                vec2 uv_r = uv + vec2(glitch, 0.0);
                vec2 uv_g = uv + vec2(-glitch * 0.5, glitch * 0.5);
                vec2 uv_b = uv + vec2(0.0, -glitch);

                // Muestreo canales desplazados
                float r = texture(tex, uv_r).r;
                float g = texture(tex, uv_g).g;
                float b = texture(tex, uv_b).b;

                f_color = vec4(r, g, b, 1.0);
            }
            '''
        )
    
    def __init__(
        self,
        size,
        opengl_context = None,
        amplitude: float = 0.02,
        frequency: float = 25.0,
        glitch_amplitude: float = 0.02,
        glitch_speed: float = 30.0
    ):
        super().__init__(
            opengl_context = opengl_context,
            size = size,
            amp = amplitude,
            freq = frequency,
            glitchAmp = glitch_amplitude,
            glitchSpeed = glitch_speed
        )

    def process(
        self,
        input: Union[moderngl.Texture, np.ndarray],
        t: float = 0.0,
    ) -> moderngl.Texture:
        """
        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        return super().process(
            input = input,
            time = t
        )


"""
TODO: This module doesn't use 't' but 'progress'
so it is not a child of 'processor.video', maybe
we should move it to be 'processor.transitions'
instead of 'processor.video.transitions'... (?)
"""
from yta_video_opengl.abstract import _OpenGLBase
from yta_validation.parameter import ParameterValidator
from typing import Union

import numpy as np
import moderngl


class _TransitionProcessorGPU(_OpenGLBase):
    """
    *Abstract class*

    *For internal use only*

    A transition between the frames of 2 videos.

    This transition is made with GPU (OpenGL).
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        """
        The code of the fragment shader.
        """
        return (
            '''
            #version 330
            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform float progress;     // 0.0 → full A, 1.0 → full B
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                // Horizontal version (right to left)
                vec2 uv_first = v_uv + vec2(-progress, 0.0);
                vec2 uv_second = v_uv + vec2(1.0 - progress, 0.0);

                vec4 color_first = texture(first_texture, uv_first);
                vec4 color_second = texture(second_texture, uv_second);

                if (uv_first.x < 0.0) {
                    output_color = color_second;
                } else if (uv_second.x > 1.0) {
                    output_color = color_first;
                } else {
                    // A and B frames are shown at the same time
                    output_color = mix(color_first, color_second, progress);
                }
            }
            '''
        )
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        **kwargs
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            **kwargs
        )

    def _prepare_input_textures(
        self
    ) -> '_OpenGLBase':
        """
        *For internal use only*

        *This method should be overwritten*

        Set the input texture variables and handlers
        we need to manage this. This method has to be
        called only once, just to set the slot for 
        the different textures we will use (and are
        registered as textures in the shader).
        """
        self.textures.add('first_texture', 0)
        self.textures.add('second_texture', 1)

        return self
    
    def process(
        self,
        first_input: Union[moderngl.Texture, np.ndarray],
        second_input: Union[moderngl.Texture, np.ndarray],
        progress: float,
        output_size: Union[tuple[int, int], None] = None,
        **kwargs
    ) -> moderngl.Texture:
        """
        Validate the parameters, set the textures map, process
        it and return the result according to the `progress`
        provided.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        ParameterValidator.validate_mandatory_instance_of('first_input', first_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_instance_of('second_input', second_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_positive_float('progress', progress, do_include_zero = True)

        textures_map = {
            'first_texture': first_input,
            'second_texture': second_input
        }

        return self._process_common(
            textures_map = textures_map,
            output_size = output_size,
            progress = progress,
            **kwargs
        )
    
# Specific implementations here below:
class SlideTransitionProcessorGPU(_TransitionProcessorGPU):
    """
    A transition between the frames of 2 videos, sliding
    from right to left.

    This transition is made with GPU (OpenGL).
    """

    # TODO: I know it is the same as in the base class
    # but I want it like that
    @property
    def fragment_shader(
        self
    ) -> str:
        """
        The code of the fragment shader.
        """
        return (
            '''
            #version 330
            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform float progress;     // 0.0 → full A, 1.0 → full B
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                // Horizontal version (slide to right)
                //vec2 uv_first = v_uv + vec2(-progress, 0.0);
                //vec2 uv_second = v_uv + vec2(1.0 - progress, 0.0);

                // Horizontal version (slide to left)
                vec2 uv_first = v_uv + vec2(progress, 0.0);
                vec2 uv_second = v_uv + vec2(-1.0 + progress, 0.0);

                vec4 color_first = texture(first_texture, uv_first);
                vec4 color_second = texture(second_texture, uv_second);

                // Horizontal version (slide to right)
                //if (uv_first.x < 0.0) {
                //    output_color = color_second;
                //} else if (uv_second.x > 1.0) {
                //    output_color = color_first;
                //} else {
                //    // A and B frames are shown at the same time
                //    output_color = mix(color_first, color_second, progress);
                //}

                // Horizontal version (slide t o left)
                if (uv_first.x > 1.0) {
                    output_color = color_second;
                } else if (uv_second.x < 0.0) {
                    output_color = color_first;
                } else {
                    output_color = mix(color_first, color_second, progress);
                }
            }
            '''
        )
    
class CrossfadeTransitionProcessorGPU(_TransitionProcessorGPU):
    """
    A transition between the frames of 2 videos,
    transforming the first one into the second one.

    This transition is made with GPU (OpenGL).
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            """
            #version 330
            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform float progress; // 0 = full A, 1 = full B
            in vec2 v_uv;
            out vec4 output_color;
            void main() {
                vec4 color_first = texture(first_texture, v_uv);
                vec4 color_second = texture(second_texture, v_uv);
                output_color = mix(color_first, color_second, progress);
            }
            """
        )
    
class DistortedCrossfadeTransitionProcessorGPU(_TransitionProcessorGPU):
    """
    A transition between the frames of 2 videos,
    transforming the first one into the second one
    with a distortion in between.

    This transition is made with GPU (OpenGL).
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            """
            #version 330
            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform float progress;   // 0.0 -> A, 1.0 -> B
            uniform float intensity;  // Distortion control
            in vec2 v_uv;
            out vec4 output_color;

            const int passes = 6;

            void main() {
                vec4 c1 = vec4(0.0);
                vec4 c2 = vec4(0.0);

                float disp = intensity * (0.5 - distance(0.5, progress));
                for (int xi=0; xi<passes; xi++) {
                    float x = float(xi) / float(passes) - 0.5;
                    for (int yi=0; yi<passes; yi++) {
                        float y = float(yi) / float(passes) - 0.5;
                        vec2 v = vec2(x, y);
                        float d = disp;
                        c1 += texture(first_texture, v_uv + d * v);
                        c2 += texture(second_texture, v_uv + d * v);
                    }
                }
                c1 /= float(passes * passes);
                c2 /= float(passes * passes);
                output_color = mix(c1, c2, progress);
            }
            """
        )
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        intensity: float = 1.0,
        **kwargs
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            intensity = intensity,
            **kwargs
        )

class AlphaPediaMaskTransitionProcessorGPU(_TransitionProcessorGPU):
    """
    A transition made by using a custom mask to
    join the 2 videos. This mask is specifically
    obtained from the AlphaPediaYT channel in which
    we upload specific masking videos.

    Both videos will be placed occupying the whole
    scene, just overlapping by using the transition
    video mask, but not moving the frame through 
    the screen like other classes do (like the
    FallingBars).
    """

    # TODO: I think I don't need a 'progress' but just
    # mix both frames as much as the alpha (or white
    # presence) tells
    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            """
            #version 330

            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform sampler2D mask_texture;

            uniform float progress;  // 0.0 → full A, 1.0 → full B
            uniform bool use_alpha_channel;   // True to use the alpha channel
            //uniform float contrast;  // Optional contrast to magnify the result

            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                vec4 first_color = texture(first_texture, v_uv);
                vec4 second_color = texture(second_texture, v_uv);
                vec4 mask_color = texture(mask_texture, v_uv);

                // Mask alpha or red?
                float mask_value = use_alpha_channel ? mask_color.a : mask_color.r;

                // Optional contrast
                //mask_value = clamp((mask_value - 0.5) * contrast + 0.5, 0.0, 1.0);
                mask_value = clamp((mask_value - 0.5) + 0.5, 0.0, 1.0);

                float t = smoothstep(0.0, 1.0, mask_value + progress - 0.5);

                output_color = mix(first_color, second_color, t);
            }
            """
        )
    
    def _prepare_input_textures(
        self
    ) -> None:
        """
        *For internal use only*

        Set the input texture variables and handlers
        we need to manage this.
        """
        self.textures.add('first_texture', 0)
        self.textures.add('second_texture', 1)
        self.textures.add('mask_texture', 2)
    
    def process(
        self,
        first_input: Union[moderngl.Texture, 'np.ndarray'],
        second_input: Union[moderngl.Texture, 'np.ndarray'],
        mask_input: Union[moderngl.Texture, 'np.ndarray'],
        progress: float,
        output_size: Union[tuple[int, int], None] = None,
        **kwargs
    ) -> moderngl.Texture:
        """
        Apply the shader to the 'input', that
        must be a frame or a texture, and return
        the new resulting texture.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        ParameterValidator.validate_mandatory_instance_of('first_input', first_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_instance_of('second_input', second_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_instance_of('mask_input', mask_input, [moderngl.Texture, np.ndarray])
        ParameterValidator.validate_mandatory_positive_float('progress', progress, do_include_zero = True)

        textures_map = {
            'first_texture': first_input,
            'second_texture': second_input,
            'mask_texture': mask_input
        }

        # TODO: There is an 'use_alpha_channel' uniform to use
        # the alpha instead of the red color of the frame value,
        # but the red is working for our AlphaPedia videos, so...

        kwargs = {
            **kwargs,
            'progress': progress
        }

        return self._process_common(
            textures_map = textures_map,
            output_size = output_size,
            **kwargs
        )
    
class CircleOpeningTransitionProcessorGPU(_TransitionProcessorGPU):
    """
    A transition between the frames of 2 videos in
    which the frames are mixed by generating a circle
    that grows from the middle to end fitting the 
    whole screen.

    This transition is made with GPU (OpenGL).
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            """
            #version 330
            #define UNIQUE_ID_{id(self)}
            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform float progress;  // 0.0 → full A, 1.0 → full B
            uniform float border_smoothness; // 0.02 is a good value

            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                // Obtain the size automatically from the texture
                vec2 output_size = vec2(textureSize(first_texture, 0));

                vec2 pos = v_uv * output_size;
                vec2 center = output_size * 0.5;

                // Distance from center
                float dist = distance(pos, center);

                // Radius of current circle
                float max_radius = length(center);
                float radius = progress * max_radius;

                vec4 first_color = texture(first_texture, v_uv);
                vec4 second_color = texture(second_texture, v_uv);

                // With smooth circle
                // TODO: Make this customizable
                float mask = 1.0 - smoothstep(radius - border_smoothness * max_radius, radius + border_smoothness * max_radius, dist);
                output_color = mix(first_color, second_color, mask);
            }
            """
        )

    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        border_smoothness: float = 0.02,
        **kwargs
    ):
        super().__init__(
            opengl_context = opengl_context,
            # TODO: Maybe 'output_size' has to be the texture size
            # by default
            output_size = output_size,
            border_smoothness = border_smoothness,
            **kwargs
        )
    
class CircleClosingTransitionProcessorGPU(_TransitionProcessorGPU):
    """
    A transition between the frames of 2 videos in
    which the frames are mixed by generating a circle
    that grows from the middle to end fitting the 
    whole screen.

    This transition is made with GPU (OpenGL).
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            """
            #version 330

            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform float progress;  // 0.0 → full A, 1.0 → full B
            uniform float border_smoothness; // 0.02 is a good value

            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                // Obtain the size automatically from the texture
                vec2 output_size = vec2(textureSize(first_texture, 0));

                vec2 pos = v_uv * output_size;
                vec2 center = output_size * 0.5;

                // Distance from center
                float dist = distance(pos, center);

                // Radius of current circle
                float max_radius = length(center);
                float radius = (1.0 - progress) * max_radius;

                vec4 first_color = texture(first_texture, v_uv);
                vec4 second_color = texture(second_texture, v_uv);

                // With smooth circle
                // TODO: Make this customizable
                float mask = smoothstep(radius - border_smoothness * max_radius, radius + border_smoothness * max_radius, dist);
                output_color = mix(first_color, second_color, mask);
            }
            """
        )

    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        border_smoothness: float = 0.02,
        **kwargs
    ):
        super().__init__(
            opengl_context = opengl_context,
            # TODO: Maybe 'output_size' has to be the texture size
            # by default
            output_size = output_size,
            border_smoothness = border_smoothness,
            **kwargs
        )
    
# TODO: This effect is not working according to
# the progress, you cannot use normal timing
class BarsFallingTransitionProcessorGPU(_TransitionProcessorGPU):
    """
    A transition between the frames of 2 videos in which
    a set of bars fall with the first video to let the
    second one be seen.

    Extracted from here:
    - https://gl-transitions.com/editor/DoomScreenTransition

    This transition is made with GPU (OpenGL).
    """

    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            """
            #version 330

            uniform sampler2D first_texture;
            uniform sampler2D second_texture;
            uniform float progress; // 0.0 → start, 1.0 → end

            uniform int number_of_bars;           
            uniform float amplitude;        // Speed
            uniform float noise;            // Extra noise [0.0, 1.0]
            uniform float frequency;        // Wave frequency
            uniform float drip_scale;       // Falling from center

            in vec2 v_uv;
            out vec4 output_color;

            // pseudo-random from integer
            float rand(int num) {
                return fract(mod(float(num) * 67123.313, 12.0) * sin(float(num) * 10.3) * cos(float(num)));
            }

            // Wave for vertical distortion
            float wave(int num) {
                float fn = float(num) * frequency * 0.1 * float(number_of_bars);
                return cos(fn * 0.5) * cos(fn * 0.13) * sin((fn + 10.0) * 0.3) / 2.0 + 0.5;
            }

            // Vertical curve to borders
            float drip(int num) {
                return sin(float(num) / float(number_of_bars - 1) * 3.141592) * drip_scale;
            }

            // Displacement for a bar
            float pos(int num) {
                float w = wave(num);
                float r = rand(num);
                float base = (noise == 0.0) ? w : mix(w, r, noise);
                return base + ((drip_scale == 0.0) ? 0.0 : drip(num));
            }

            void main() {
                int bar = int(v_uv.x * float(number_of_bars));

                float scale = 1.0 + pos(bar) * amplitude;
                float phase = progress * scale;
                float pos_y = v_uv.y;

                vec2 p;
                vec4 color;

                if (phase + pos_y < 1.0) {
                    // Frame A is visible
                    p = vec2(v_uv.x, v_uv.y + mix(0.0, 1.0, phase));
                    color = texture(first_texture, p);
                } else {
                    // Frame B is visible
                    color = texture(second_texture, v_uv);
                }

                output_color = color;
            }
            """
        )

    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        # TODO: Review this
        output_size: tuple[int, int] = (1920, 1080),
        number_of_bars: int = 30,
        amplitude: float = 2.0,
        noise: float = 0.1, # [0.0, 1.0]
        frequency: float = 0.5,
        drip_scale: float = 0.5,
        **kwargs
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            number_of_bars = number_of_bars,
            amplitude = amplitude,
            noise = noise,
            frequency = frequency,
            drip_scale = drip_scale,
            **kwargs
        )

    
"""
Note for the developer:

Here below you have a shader that allows you
to create more slide transitions (vertical,
diagonal) but have to be refactored because
the mixing part is not working properly
according to the position. The code was made
for an horizontal slide but has to be adapted
to the other movements.

Code here below:

#version 330

// FRAGMENT SHADER — Slide horizontal
uniform sampler2D texA;
uniform sampler2D texB;
uniform float progress;     // 0.0 → full A, 1.0 → full B

in vec2 frag_uv;
out vec4 frag_color;

void main() {
    // Horizontal version (right to left)
    vec2 uvA = frag_uv + vec2(-progress, 0.0);
    vec2 uvB = frag_uv + vec2(1.0 - progress, 0.0);
    
    // Horizontal version (left to right)
    //vec2 uvA = frag_uv + vec2(progress, 0.0);
    //vec2 uvB = frag_uv + vec2(-1.0 + progress, 0.0);

    // Vertical version (top to bottom)
    // TODO: We need to adjust the color mixin
    // to make it fit the type of transition
    //vec2 uvA = frag_uv + vec2(0.0, -progress);
    //vec2 uvB = frag_uv + vec2(0.0, 1.0 - progress);

    // Diagonal version (top left to bottom right)
    //vec2 uvA = frag_uv + vec2(-progress, -progress);
    //vec2 uvB = frag_uv + vec2(1.0 - progress, 1.0 - progress);

    vec4 colorA = texture(texA, uvA);
    vec4 colorB = texture(texB, uvB);

    if (uvA.x < 0.0) {
        frag_color = colorB;
    } else if (uvB.x > 1.0) {
        frag_color = colorA;
    } else {
        // A and B frames are shown at the same time
        frag_color = mix(colorA, colorB, progress);
    }
}
"""


"""
Module for the effects and nodes that are made by
putting different nodes together.
"""
from yta_editor_nodes_gpu.processor import _NodeComplexGPU
from typing import Union

import moderngl


# TODO: This is special as it is not a node itself
class DisplayOverAtNodeComplexGPU(_NodeComplexGPU):
    """
    The overlay input is positioned with the given position,
    rotation and size, and then put as an overlay of the
    also given base input.

    Information:
    - The scene size is `(1920, 1080)`, so provide the
    `position` parameter according to it, where it is
    representing the center of the texture.
    - The `rotation` is in degrees, where `rotation=90`
    means rotating 90 degrees to the right.
    - The `size` parameter must be provided according to
    the previously mentioned scene size `(1920, 1080)`.

    TODO: This has no inheritance, is special and we need
    to be able to identify it as a valid one.
    """

    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        **kwargs
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            **kwargs
        )

    def process(
        self,
        base_input: Union['moderngl.Texture', 'np.ndarray', None],
        overlay_input: Union['moderngl.Texture', 'np.ndarray'],
        position: tuple[int, int] = (1920 / 2, 1080 / 2),
        size: tuple[int, int] = (1920 / 2, 1080 / 2),
        rotation: int = 0
    ):
        """
        By default, the texture overlayed will be displayed in
        the center of the scene, with half of the scene size
        and no rotation.
        """
        from yta_editor_nodes_gpu.processor import DisplacementWithRotationNodeCompositorGPU
        from yta_editor_nodes_gpu.processor.blender import AlphaBlenderGPU

        displacement_node_processor = DisplacementWithRotationNodeCompositorGPU(
            opengl_context = self._opengl_context,
            output_size = self._output_size
        )
        output = displacement_node_processor.process(
            #base_input = background_as_numpy,
            base_input = None,
            #overlay_input = background_as_numpy,
            overlay_input = overlay_input,
            output_size = None,
            position = position,
            size = size,
            rotation = rotation
            #rotation = 0.785398 # 0.785398 = 45deg
        )

        if base_input is not None:
            #  TODO: Add just as an overlay
            blender = AlphaBlenderGPU(
                opengl_context = self._opengl_context,
                output_size = self._output_size
            )
            output = blender.process(
                # We don't need to care about the size because OpenGL
                # handles it
                base_input = base_input,
                overlay_input = output,
                mix_weight = 1.0,
                blend_strength = 1.0
            )

        return output


"""
The nodes that modify inputs by using static parameters
and not 't' time moments, with GPU.
"""
from yta_video_opengl.abstract import _OpenGLBase
from yta_validation.parameter import ParameterValidator
from typing import Union
from abc import abstractmethod

import moderngl


# TODO: Move this abstract classes and reorganize
# the modules because we have only one 'processor'
# class but the module is called like that...
class _NodeProcessorGPU(_OpenGLBase):
    """
    *Singleton class*

    *For internal use only*
    
    Class to represent a node processor that uses GPU
    to transform the input.

    The basic class of a node to manipulate frames as
    opengl textures. This node will process the frame
    as an input texture and will generate also a
    texture as the output.

    Nodes can be chained and the result from one node
    can be applied on another node.
    """

    # TODO: Do we really need this class just to add
    # one step on the hierarchy? It is to differentiate
    # between any class that uses OpenGL on the base
    # from the ones that are nodes to process and to
    # be added to a general NodeProcessor class
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        **kwargs
    ):
        """
        Provide all the variables you want to be initialized
        as uniforms at the begining for the global OpenGL
        animation in the `**kwargs`.

        The `output_size` is the size (width, height) of the
        texture that will be obtained as result. This size
        can be modified when processing a specific input, but
        be consider the cost of resources of modifying the 
        size, that will regenerate the output texture.
        """
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            **kwargs
        )

    def __reinit__(
        self,
        opengl_context: Union[moderngl.Context, None] = None,
        output_size: Union[tuple[int, int], None] = None,
        **kwargs
    ):
        # TODO: We ignore the 'opengl_context' as we don't need
        # to reset it
        super().__reinit__(
            output_size = output_size,
            **kwargs
        )

class _NodeComplexGPU:
    """
    A node that is made by applying different nodes. It
    is different than the other nodes because it needs
    to import different nodes to process the input(s).

    TODO: This is, by now, just a class to identify these
    new type of nodes.
    """

    def __init__(
        self,
        opengl_context: Union['moderngl.Context', None],
        output_size: Union[tuple[int, int], None],
        **kwargs
    ):
        self._opengl_context: Union['moderngl.Context', None] = opengl_context
        self._output_size: Union[tuple[int, int], None] = output_size
        self.kwargs = kwargs

    @abstractmethod
    def process(
        self,
        # If None it must be an empty alpha texture
        base_input: Union[moderngl.Texture, 'np.ndarray', None],
        output_size: Union[tuple[int, int], None],
        **kwargs
    ) -> moderngl.Texture:
        pass


# TODO: Testing this one, to position different in
# the scene
class _NodeCompositorGPU(_OpenGLBase):
    """
    *Singleton class*

    *For internal use only*
    
    Class to represent a node processor that uses GPU
    to composite the input into the scene, by rotating,
    positioning it, etc.

    This node will process the frame as an input
    texture and will generate also a texture as the
    output.

    Nodes can be chained and the result from one node
    can be applied on another node.
    """

    @property
    def vertex_shader(
        self
    ) -> str:
        """
        The code of the vertex shader. This is, by default,
        a rectangle made by 2 triangles that will be placed
        in the specific position, with the size provided and
        with the also given rotation.
        """
        return (
            '''
            #version 330

            in vec2 in_vert;        // Quad vertices [-1.0, 1.0]
            in vec2 in_texcoord;    // UV Coordinates: (0, 0) a (1, 1)
            out vec2 v_uv;

            uniform vec2 position;  // Center of texture
            uniform vec2 size;      // Size of texture (w, h)
            uniform float rotation; // Rotation (in degrees)

            // Resolution of the viewport
            // TODO: Turn this into a uniform with default value
            const vec2 resolution = vec2(1920.0, 1080.0);

            vec2 rotate_around_center(vec2 p, float rotation) {
                float radians = radians(rotation);
                float s = sin(radians);
                float c = cos(radians);
                mat2 rot = mat2(c, -s, s, c);
                return rot * p;
            }

            void main() {
                // Local coordinates but in [-0.5, 0.5]
                vec2 local = in_vert * 0.5;

                vec2 scaled = local * size;
                vec2 rotated = rotate_around_center(scaled, rotation);
                vec2 pos_pixels = position + rotated;

                // Pixels to normalized coordinates [-1.0, 1.0]
                vec2 ndc = (pos_pixels / resolution) * 2.0 - 1.0;

                gl_Position = vec4(ndc, 0.0, 1.0);
                v_uv = in_texcoord;
            }
            '''
        )
    
    @property
    def fragment_shader(
        self
    ) -> str:
        return (
            '''
            #version 330
            in vec2 v_uv;
            out vec4 output_color;
            uniform sampler2D base_texture;
            uniform sampler2D overlay_texture;
            uniform bool do_use_transparent_base_texture;

            void main() {
                vec4 base_color;

                if (do_use_transparent_base_texture) {
                    base_color = vec4(0.0, 0.0, 0.0, 0.0);
                } else {
                    //base_color = vec4(1.0, 1.0, 1.0, 1.0);
                    base_color = texture(base_texture, v_uv);
                }

                vec4 overlay_color = texture(overlay_texture, v_uv);
                output_color = mix(base_color, overlay_color, overlay_color.a);
                //output_color = vec4(v_uv, 0.0, 1.0);
            }
            '''
        )
    
    def _prepare_input_textures(
        self
    ) -> '_OpenGLBase':
        """
        *For internal use only*

        Set the input texture variables and handlers
        we need to manage this. This method has to be
        called only once, just to set the slot for 
        the different textures we will use (and are
        registered as textures in the shader).
        """
        self.textures.add('base_texture', 0)
        self.textures.add('overlay_texture', 1)

        return self
    
    def process(
        self,
        # If None it must be an empty alpha texture
        base_input: Union[moderngl.Texture, 'np.ndarray', None],
        overlay_input: Union[moderngl.Texture, 'np.ndarray'],
        output_size: Union[tuple[int, int], None],
        #output_size: Union[tuple[int, int], None] = None,
        # TODO: We need the value as pixels in (1920, 1080)
        position: tuple[int, int],
        size: tuple[int, int],
        rotation: int,
        **kwargs
    ) -> moderngl.Texture:
        """
        Mix the `processed_input` with the 
        `original_input` as the `selecction_mask_input`
        says.

        The `position` and `size` parameters must be in
        pixels (in a scene size of `(1920, 1080)`). The
        rotation must be an `int` value representing the
        degrees.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        ParameterValidator.validate_instance_of('base_input', base_input, [moderngl.Texture, 'ndarray'])
        ParameterValidator.validate_mandatory_instance_of('overlay_input', overlay_input, [moderngl.Texture, 'ndarray'])

        do_use_transparent_base_texture = base_input is None

        textures_map = {}

        if base_input is not None:
            # TODO: The size will be obtained from the first texture
            # dynamically, but if we don't provide the base texture,
            # it will be created as a completely transparent one,
            # that doesn't have any size in the 'textures_map' var
            textures_map['base_texture'] = base_input
        
        textures_map = textures_map | {
            # TODO: I think the 'base_input' should be forced to
            # be a completely transparent texture here
            'overlay_texture': overlay_input
        }

        return self._process_common(
            textures_map = textures_map,
            output_size = output_size,
            position = position,
            size = size,
            rotation = rotation,
            do_use_transparent_base_texture = do_use_transparent_base_texture,
            **kwargs
        )

class SelectionMaskProcessorGPU(_NodeProcessorGPU):
    """
    Class to use a mask selection (from which we will
    read the red color to determine if the pixel must
    be applied or not) to apply the `processed_texture`
    on the `original_texture`.

    If the selection mask is completely full, the
    result will be the `processed_texture`.
    """
    
    @property
    def fragment_shader(
        self
    ):
        return (
            '''
            #version 330

            uniform sampler2D original_texture;
            uniform sampler2D processed_texture;
            // White = apply, black = ignore
            uniform sampler2D selection_mask_texture;
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                vec4 original_color = texture(original_texture, v_uv);
                vec4 processed_color = texture(processed_texture, v_uv);
                // We use the red as the value
                float mask = texture(selection_mask_texture, v_uv).r; 

                output_color = mix(original_color, processed_color, mask);
            }
            '''
        )
    
    def _prepare_input_textures(
        self
    ) -> '_OpenGLBase':
        """
        *For internal use only*

        Set the input texture variables and handlers
        we need to manage this. This method has to be
        called only once, just to set the slot for 
        the different textures we will use (and are
        registered as textures in the shader).
        """
        self.textures.add('original_texture', 0)
        self.textures.add('processed_texture', 1)
        self.textures.add('selection_mask_texture', 2)

        return self

    def process(
        self,
        original_input: Union[moderngl.Texture, 'np.ndarray'],
        processed_input: Union[moderngl.Texture, 'np.ndarray'],
        selection_mask_input: Union[moderngl.Texture, 'np.ndarray'],
        output_size: Union[tuple[int, int], None] = None,
        **kwargs
    ) -> moderngl.Texture:
        """
        Mix the `processed_input` with the 
        `original_input` as the `selecction_mask_input`
        says.

        You can provide any additional parameter
        in the **kwargs, but be careful because
        this could overwrite other uniforms that
        were previously set.

        We use and return textures to maintain
        the process in GPU and optimize it.
        """
        ParameterValidator.validate_mandatory_instance_of('original_input', original_input, [moderngl.Texture, 'ndarray'])
        ParameterValidator.validate_mandatory_instance_of('processed_input', processed_input, [moderngl.Texture, 'ndarray'])
        ParameterValidator.validate_mandatory_instance_of('selection_mask_input', selection_mask_input, [moderngl.Texture, 'ndarray'])

        textures_map = {
            'original_texture': original_input,
            'processed_texture': processed_input,
            'selection_mask_texture': selection_mask_input
        }

        return self._process_common(
            textures_map = textures_map,
            output_size = output_size,
            **kwargs
        )
    
# TODO: Just for testing, at least by now
class ColorContrastNodeProcessorGPU(_NodeProcessorGPU):
    """
    Node processor to test the color contrast
    change.

    Multiplicas la distancia al gris medio.
    Así, los claros se vuelven más claros y los oscuros más oscuros.
    Es la forma más simple (y lineal) de aumentar contraste sin alterar el brillo global.

    TODO: Improve this, its just temporary
    """

    @property
    def fragment_shader(
        self
    ):
        return (
            '''
            #version 330

            uniform sampler2D base_texture;
            uniform float factor;
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                vec4 color = texture(base_texture, v_uv);
                vec3 mean = vec3(0.5);
                color.rgb = (color.rgb - mean) * factor + mean;
                output_color = vec4(clamp(color.rgb, 0.0, 1.0), color.a);
            }
            '''
        )
    
    def __init__(
        self,
        opengl_context: Union[moderngl.Context, None],
        output_size: tuple[int, int],
        factor: float = 1.5,
        **kwargs
    ):
        super().__init__(
            opengl_context = opengl_context,
            output_size = output_size,
            factor = factor,
            **kwargs
        )

    def __reinit__(
        self,
        opengl_context: Union[moderngl.Context, None] = None,
        output_size: Union[tuple[int, int], None] = None,
        factor: Union[float, None] = None,
        **kwargs
    ):
        super().__reinit__(
            opengl_context = opengl_context,
            output_size = output_size,
            # TODO: Maybe this one has to be None or added
            # dynamically to kwargs only if not None
            factor = factor,
            **kwargs
        )

class BlackAndWhiteNodeProcessorGPU(_NodeProcessorGPU):
    """
    Node processor to test the black and white
    effect.

    TODO: Improve this, its just temporary
    """

    @property
    def fragment_shader(
        self
    ):
        return (
            '''
            #version 330

            uniform sampler2D base_texture;
            uniform float factor;
            in vec2 v_uv;
            out vec4 output_color;

            void main() {
                vec4 color = texture(base_texture, v_uv);
                // Luminance Rec.709
                float gray = dot(color.rgb, vec3(0.2126, 0.7152, 0.0722));
                output_color = vec4(vec3(gray), color.a);
            }
            '''
        )
    
# TODO: Move to a 'compositor' folder (?)
class DisplacementWithRotationNodeCompositorGPU(_NodeCompositorGPU):
    """
    The frame, but moving and rotating over other frame.
    """

    pass